"""
Statement result types.
"""
import sys
from abc import ABC, abstractmethod
from typing import List
from tabulate import tabulate


class StatementResult(ABC):
    """
    A statement result.
    """

    @abstractmethod
    def __str__(self):
        """
        This query result as a string. Used for printing it.
        """


class DMLResult(StatementResult):
    """
    A DML/DDL result.
    """

    def __init__(self, number_updated: int):
        self.number_updated = number_updated

    def __str__(self):
        return f"Number of record updated: {self.number_updated}"


class QueryResult(StatementResult):
    """
    A query result.
    """

    def __init__(self, headers: List[str], rows: List[List[str]]):
        self.headers = headers
        self.rows = rows

    def __str__(self):
        tablefmt = "psql"
        headers = self.headers
        if not sys.stdin.isatty():
            tablefmt = "tsv"
            headers = []
        return tabulate(self.rows, headers=headers, tablefmt=tablefmt)



"""
Settings module for the cli.
"""
import os
import logging
import yaml
import boto3
from rdsline.connections import NoopConnection
from rdsline.connections.rds_secretsmanager import RDSSecretsManagerConnection

DEFAULT_CONFIG_FILE = os.path.expanduser("~/.rdsline")


def _get_region(cluster_arn: str):
    arn_parts = cluster_arn.split(":")
    return arn_parts[3]


def _client_provider(profile: str, region: str):
    logging.debug("Setting aws credentials profile to %s - region: %s", profile, region)
    session = boto3.Session(profile_name=profile)
    return session.client("rds-data", region_name=region)


def from_file(file: str, client_provider=_client_provider):
    """
    Reads settings from a file.
    """
    logging.debug("Reading configuration file from: %s", file)
    settings = {}
    with open(file, "r", encoding="utf-8") as stream:
        settings = yaml.safe_load(stream)
    logging.debug("Settings: %s", settings)
    if settings["type"] != "rds-secretsmanager":
        raise Exception(f"Unsupported database connection type: {settings['type']}")
    region = _get_region(settings["cluster_arn"])
    profile = settings["credentials"]["profile"]
    client = client_provider(profile, region)
    return RDSSecretsManagerConnection(
        settings["cluster_arn"], settings["secret_arn"], settings["database"], client
    )


def from_args(args):
    """
    Reads settings from cli args.
    """
    if args.config is not None:
        return from_file(args.config)
    if os.path.exists(DEFAULT_CONFIG_FILE):
        return from_file(DEFAULT_CONFIG_FILE)
    logging.debug("No config. Set to null connector")
    return NoopConnection()



"""
Connection abstract class.
"""
from abc import ABC, abstractmethod
from rdsline.results import StatementResult


class Connection(ABC):
    """
    Connection for a database.
    """

    @abstractmethod
    def execute(self, sql: str) -> StatementResult:
        """
        Executes a query using this connection.
        """

    @abstractmethod
    def is_executable(self) -> bool:
        """
        Returns if this connection is executable (i.e. all settings are correctly set)
        """

    @abstractmethod
    def __str__(self):
        """
        Returns this connection as a string. Meant to be displayed.
        """


class NoopConnection(Connection):
    """
    No op connection.
    """

    def __init__(self):
        pass

    def is_executable(self) -> bool:
        return False

    def __str__(self):
        return "No connection settings. Maybe you need to run .config"

    def execute(self, _: str) -> StatementResult:
        raise NotImplementedError("Execute in No op")



"""
The main entry point for rdsline.
"""
import logging
import argparse
import readline  # pylint: disable=unused-import
import os
import sys
from platform import python_version_tuple
from typing import List
from rdsline import settings
from rdsline.version import VERSION
from rdsline.connections import Connection


if python_version_tuple() < ("3", "7", "8"):
    raise Exception("We don't support Python < 3.7.8")


def _help():
    return "\n".join(
        [
            ".quit - quits the REPL",
            ".config <config_file> - sets new connection settings from a file",
            ".show - displays current connection settings",
            ".debug - toggle debugging information",
        ]
    )


def _read(prompt: str) -> str:
    try:
        line = input(f"{prompt}")
        if line == ".quit":
            sys.exit(0)
        return line
    except EOFError:
        sys.exit(0)


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="The RDS REPL v" + VERSION)
    parser.add_argument(
        "--config", type=str, required=False, help="Config file to read settings from"
    )
    parser.add_argument(
        "--debug", required=False, action="store_true", help="Turn debugging information on."
    )
    return parser.parse_args()


class DebugCommand:
    """
    Toggles debug logging on/off.
    """

    def __init__(self, is_debug: bool):
        self.is_debug = is_debug
        logging.basicConfig(level=logging.DEBUG if is_debug else logging.WARN)

    def __call__(self, _):
        self.is_debug = not self.is_debug
        logging.getLogger().setLevel(logging.DEBUG if self.is_debug else logging.WARN)
        return "Debugging is " + ("ON" if self.is_debug else "OFF")


class ConfigCommand:
    """
    Config command.
    """

    def __init__(self, connection: Connection):
        self.connection = connection

    def __call__(self, args: List[str]) -> str:
        if len(args) != 2:
            return "ERROR: Expecting config file"
        (_, config_file) = args
        self.connection = settings.from_file(os.path.expanduser(config_file))
        return str(self.connection)


def main():
    """
    The main entry point for the program.
    """
    args = _parse_args()
    config = ConfigCommand(settings.from_args(args))
    commands = {
        ".help": lambda _: _help(),
        ".show": lambda _: str(config.connection),
        ".debug": DebugCommand(args.debug),
        ".config": config,
    }
    buffer = ""
    default_prompt = ""
    if sys.stdin.isatty():
        default_prompt = "> "
        print("The RDS REPL v" + VERSION)
        print("Type .help for help")
    prompt = default_prompt
    while True:
        line = _read(prompt)
        if line and line[0] == ".":
            args = line.split(" ")
            if args[0] in commands:
                print(commands[args[0]](args))
        elif line.endswith(";") or line == "":
            buffer += line
            if not config.connection.is_executable():
                print(config.connection)
                buffer = ""
                prompt = default_prompt
                continue
            try:
                print(config.connection.execute(buffer))
            except Exception as ex:  # pylint: disable=broad-except
                print(f"Error: {str(ex)}")
            finally:
                buffer = ""
                prompt = default_prompt
        else:
            buffer += line + " "
            prompt = "|"


if __name__ == "__main__":
    main()



"""
Connection for RDS with a Secrets Manager secret.
"""
import logging
from typing import Any
from rdsline.connections import Connection
from rdsline.results import DMLResult, QueryResult, StatementResult


def _to_string(val: Any) -> str:
    known_types = {
        "stringValue": lambda v: v["stringValue"],
        "booleanValue": lambda v: str(v["booleanValue"]),
        "doubleValue": lambda v: str(v["doubleValue"]),
        "longValue": lambda v: str(v["longValue"]),
        "blobValue": lambda v: "BLOB",
        "arrayValue": lambda v: "ARRAY",
    }
    if "isNull" in val and val["isNull"]:
        return "NULL"
    for (type_name, converter) in known_types.items():
        if type_name in val:
            return converter(val)
    return "UNKNOWN"


def _to_result(response) -> StatementResult:
    if "columnMetadata" not in response:
        # Assumed to be a DML/DDL operation
        if "numberOfRecordsUpdated" in response:
            return DMLResult(response["numberOfRecordsUpdated"])
    headers = [col["name"] for col in response["columnMetadata"]]
    records = [[_to_string(value) for value in row] for row in response["records"]]
    return QueryResult(headers, records)


class RDSSecretsManagerConnection(Connection):
    """
    Connection for RDS with secretsmanager.
    """

    def __init__(self, cluster_arn: str, secret_arn: str, database: str, client):
        self.cluster_arn = cluster_arn
        self.secret_arn = secret_arn
        self.database = database
        self.client = client

    def execute(self, sql: str) -> StatementResult:
        logging.debug("Executing query: %s", sql)
        response = self.client.execute_statement(
            resourceArn=self.cluster_arn,
            database=self.database,
            secretArn=self.secret_arn,
            includeResultMetadata=True,
            sql=sql,
        )
        logging.debug("Got response: %s", response)
        return _to_result(response)

    def is_executable(self) -> bool:
        return (
            self.cluster_arn is not None
            and self.secret_arn is not None
            and self.database is not None
            and self.client is not None
        )

    def __str__(self):
        return "\n".join(
            [
                "Type: rds-secretsmanager",
                f"Cluster arn: {self.cluster_arn}",
                f"Secret arn: {self.secret_arn}",
                f"Database: {self.database}",
            ]
        )



from setuptools import setup
from rdsline.version import VERSION

setup(
    name='rdsline',
    version=VERSION,
    author='Herman Polloni',
    author_email='hpolloni@gmail.com',
    description='The RDS REPL',
    entry_points = {
        'console_scripts': ['rdsline=rdsline.cli:main']
    },
    packages=['rdsline', 'rdsline.connections'],
    license="MIT",
    url='https://github.com/hpolloni/rdsline',
    long_description='RDS REPL',
    install_requires=['pyyaml', 'boto3', 'tabulate']
)
